'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
import importlib.util


def apply(input_path, validation_path, parameters=None):
    if not importlib.util.find_spec("jsonschema"):
        raise Exception(
            "please install jsonschema in order to validate a JSONOCEL file."
        )

    import json
    import jsonschema
    from jsonschema import validate

    if parameters is None:
        parameters = {}

    file_content = json.load(open(input_path, "rb"))
    schema_content = json.load(open(validation_path, "rb"))
    try:
        validate(instance=file_content, schema=schema_content)
        return True
    except jsonschema.exceptions.ValidationError as err:
        return False



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
import uuid
from typing import Optional, Dict, Any, Set, Tuple
from graphviz import Digraph, Graph
from enum import Enum
from pm4py.util import exec_utils
import tempfile
from pm4py.util import vis_utils, constants
from pm4py.objects.ocel.obj import OCEL


class Parameters(Enum):
    FORMAT = "format"
    BGCOLOR = "bgcolor"
    RANKDIR = "rankdir"
    DIRECTED = "directed"
    ENABLE_GRAPH_TITLE = "enable_graph_title"
    GRAPH_TITLE = "graph_title"


def ot_to_color(ot: str) -> str:
    ot = int(hash(ot))
    num = []
    while len(num) < 6:
        num.insert(0, ot % 16)
        ot = ot // 16
    ret = "#" + "".join([vis_utils.get_corr_hex(x) for x in num])
    return ret


def apply(
    ocel: OCEL,
    graph: Set[Tuple[str, str]],
    parameters: Optional[Dict[Any, Any]] = None,
) -> Digraph:
    """
    Visualizes an object graph

    Parameters
    -------------
    ocel
        Object-centric event log
    graph
        Object graph
    parameters
        Variant-specific parameters:
            - Parameters.FORMAT => the format of the visualization ("png", "svg", ...)
            - Parameters.BGCOLOR => the background color
            - Parameters.RANKDIR => the rank direction (LR = left-right, TB = top-bottom)
            - Parameters.DIRECTED => boolean value (draws a directed or undirected graph)

    Returns
    -------------
    gviz
        Graphviz object
    """
    if parameters is None:
        parameters = {}

    image_format = exec_utils.get_param_value(
        Parameters.FORMAT, parameters, "png"
    )
    bgcolor = exec_utils.get_param_value(
        Parameters.BGCOLOR, parameters, "transparent"
    )
    rankdir = exec_utils.get_param_value(
        Parameters.RANKDIR, parameters, constants.DEFAULT_RANKDIR_GVIZ
    )
    directed = exec_utils.get_param_value(
        Parameters.DIRECTED, parameters, True
    )

    enable_graph_title = exec_utils.get_param_value(
        Parameters.ENABLE_GRAPH_TITLE,
        parameters,
        constants.DEFAULT_ENABLE_GRAPH_TITLES,
    )
    graph_title = exec_utils.get_param_value(
        Parameters.GRAPH_TITLE, parameters, "Object-Centric Graph"
    )

    filename = tempfile.NamedTemporaryFile(suffix=".gv")
    filename.close()

    if directed:
        viz = Digraph(
            "ograph",
            filename=filename.name,
            engine="dot",
            graph_attr={"bgcolor": bgcolor},
        )
    else:
        viz = Graph(
            "ograph",
            filename=filename.name,
            engine="dot",
            graph_attr={"bgcolor": bgcolor},
        )
    viz.attr("node", shape="ellipse", fixedsize="false")

    if enable_graph_title:
        viz.attr(
            label='<<FONT POINT-SIZE="20">' + graph_title + "</FONT>>",
            labelloc="top",
        )

    ob_type = (
        ocel.objects.groupby(ocel.object_id_column)
        .first()[ocel.object_type_column]
        .to_dict()
    )

    nodes = set(x[0] for x in graph).union(set(x[1] for x in graph))
    nodes_dict = {}
    for n in nodes:
        v = str(uuid.uuid4())
        nodes_dict[n] = v
        viz.node(
            v,
            label=n,
            fontcolor=ot_to_color(ob_type[n]),
            color=ot_to_color(ob_type[n]),
        )

    for e in graph:
        viz.edge(nodes_dict[e[0]], nodes_dict[e[1]])

    viz.attr(rankdir=rankdir)
    viz.format = image_format.replace("html", "plain-ext")

    return viz



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from enum import Enum

from pm4py.objects.dfg.importer.variants import classic
from pm4py.util import exec_utils


class Variants(Enum):
    CLASSIC = classic


DEFAULT_VARIANT = Variants.CLASSIC


def apply(file_path, variant=DEFAULT_VARIANT, parameters=None):
    """
    Import a DFG (along with the start and end activities)

    Parameters
    --------------
    file_path
        Path of the DFG file
    variant
        Variant of the importer, possible values:
            - Variants.CLASSIC: importing from a .dfg file
    parameters
        Possible parameters of the algorithm

    Returns
    --------------
    dfg
        DFG
    start_activities
        Start activities
    end_activities
        End activities
    """
    return exec_utils.get_variant(variant).apply(
        file_path, parameters=parameters
    )


def deserialize(dfg_string, variant=DEFAULT_VARIANT, parameters=None):
    """
    Import a DFG from a binary/textual string

    Parameters
    --------------
    dfg_string
        DFG represented as a string in the .dfg format
    variant
        Variant of the importer, possible values:
            - Variants.CLASSIC: importing from a .dfg file
    parameters
        Possible parameters of the algorithm

    Returns
    --------------
    dfg
        DFG
    start_activities
        Start activities
    end_activities
        End activities
    """
    return exec_utils.get_variant(variant).import_dfg_from_string(
        dfg_string, parameters=parameters
    )



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from threading import Thread, Lock
import pygetwindow as gw
from pynput import mouse, keyboard
from datetime import datetime
from pynput.keyboard import Key
from pm4py.objects.log.obj import Event
import time
import psutil
import win32process
import os
from PIL import ImageGrab


class WindowsEventLogger(Thread):
    def __init__(
        self, event_stream, context_keys=None, screenshots_folder=None
    ):
        """
        Keyboard and mouse click recorder for Windows. This is useful for task mining purposes.

        Parameters
        ----------
        event_stream
            Event stream that should be used to record the events
        context_keys
            (if provided) list of keyboard keys to be considered context-switching (default:
            {Key.tab, Key.enter, Key.esc, Key.f1, Key.f2, Key.f3, Key.f4, Key.f5, Key.f6, Key.f7,
                                 Key.f8, Key.f9, Key.f10, Key.f11, Key.f12, Key.f13, Key.f14, Key.f15, Key.f16, Key.f16,
                                 Key.f17, Key.f18, Key.f19, Key.f20, Key.home, Key.end}
            )
        screenshots_folder
            (if provided) folder in which the screenshots shall be saved

        Minimum Viable Example:

            from pm4py.streaming.stream.live_event_stream import LiveEventStream
            from pm4py.streaming.connectors.windows.click_key_logger import WindowsEventLogger
            from pm4py.streaming.util.event_stream_printer import EventStreamPrinter
            import time

            stream = LiveEventStream()
            wel = WindowsEventLogger(stream, screenshots_folder="output")
            printer = EventStreamPrinter()
            stream.register(printer)
            stream.start()
            wel.start()

            print("listening")

            # listen only for 5 seconds
            time.sleep(5)

            wel.stop()
            stream.stop()

            print("stopped")
        """
        Thread.__init__(self)

        self.event_stream = event_stream
        self.lock = Lock()
        self.context_keys = context_keys
        if self.context_keys is None:
            self.context_keys = {
                Key.tab,
                Key.enter,
                Key.esc,
                Key.f1,
                Key.f2,
                Key.f3,
                Key.f4,
                Key.f5,
                Key.f6,
                Key.f7,
                Key.f8,
                Key.f9,
                Key.f10,
                Key.f11,
                Key.f12,
                Key.f13,
                Key.f14,
                Key.f15,
                Key.f16,
                Key.f16,
                Key.f17,
                Key.f18,
                Key.f19,
                Key.f20,
                Key.home,
                Key.end,
            }

        self.screenshots_folder = screenshots_folder
        if self.screenshots_folder is not None:
            if not os.path.exists(self.screenshots_folder):
                os.mkdir(self.screenshots_folder)

        self.mouse_listener = mouse.Listener(on_click=self.on_click)
        self.keyboard_listener = keyboard.Listener(
            on_release=self.on_key_release
        )

    def run(self):
        self.mouse_listener.start()
        self.keyboard_listener.start()

        self.mouse_listener.join()
        self.keyboard_listener.join()

    def stop(self):
        self.mouse_listener.stop()
        self.keyboard_listener.stop()

    def get_process_name(self, hwnd):
        """Fetch the process name using the window handle."""
        try:
            # Get the process ID associated with the handle
            hwnd = gw.getActiveWindow()._hWnd
            threadid, pid = win32process.GetWindowThreadProcessId(hwnd)
            process = psutil.Process(pid)
            return process.name()
        except (psutil.NoSuchProcess, IndexError):
            return None

    def record(
        self,
        windows_title,
        process_name=None,
        x=None,
        y=None,
        button=None,
        key=None,
    ):
        self.lock.acquire()

        process_name = str(process_name) if process_name is not None else ""
        x = str(x) if x is not None else ""
        y = str(y) if y is not None else ""
        button = str(button) if button is not None else ""
        key = str(key) if key is not None else ""

        timestamp = time.time()
        dt_object = datetime.utcfromtimestamp(timestamp)

        image_path = ""
        img = None

        if self.screenshots_folder is not None:
            image_path = os.path.join(
                self.screenshots_folder,
                "screenshot_%s.png"
                % (
                    dt_object.strftime("%Y%m%d_%H%M%S")
                    + "_"
                    + dt_object.strftime("%f")[:3]
                ),
            )

        event = Event(
            {
                "timestamp": "NONE",
                "windows_title": windows_title,
                "process_name": process_name,
                "x": x,
                "y": y,
                "button": button,
                "key": key,
                "screenshot": image_path,
            }
        )
        self.event_stream.append(event)

        if image_path:
            img = ImageGrab.grab(all_screens=True)

        self.lock.release()

        if image_path:
            img.save(image_path, "PNG")

    def on_click(self, x, y, button, pressed):
        if pressed:
            # Introduce a short delay to allow window context switch
            time.sleep(0.1)

            # Get the title of the now active window
            window_title = gw.getActiveWindow().title

            # Get the process name using the window handle
            process_name = self.get_process_name(gw.getActiveWindow()._hWnd)

            self.record(
                window_title,
                process_name=process_name,
                x=x,
                y=y,
                button=button,
            )

    def on_key_release(self, key):
        # Check if the released key is in the list
        if key in self.context_keys:
            # Get the title of the active window
            window_title = gw.getActiveWindow().title

            # Get the process name using the window handle
            process_name = self.get_process_name(gw.getActiveWindow()._hWnd)

            self.record(window_title, process_name=process_name, key=key)



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from enum import Enum
from pm4py.util import exec_utils
from pm4py.streaming.algo.conformance.footprints.variants import classic


class Variants(Enum):
    CLASSIC = classic


def apply(footprints, variant=Variants.CLASSIC, parameters=None):
    """
    Gets a footprints conformance checking object

    Parameters
    --------------
    footprints
        Footprints object (calculated from an entire log, from a process tree ...)
    variant
        Variant of the algorithm. Possible values: Variants.CLASSIC
    parameters
        Parameters of the algorithm

    Returns
    --------------
    fp_check_obj
        Footprints conformance checking object
    """
    return exec_utils.get_variant(variant).apply(
        footprints, parameters=parameters
    )



'''
    PM4Py â€“ A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschrÃ¤nkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
import pandas as pd
from typing import Optional, Dict, Any
from pm4py.util import constants, xes_constants, exec_utils
from enum import Enum


class Parameters(Enum):
    ACTIVITY_KEY = constants.PARAMETER_CONSTANT_ACTIVITY_KEY
    TIMESTAMP_KEY = constants.PARAMETER_CONSTANT_TIMESTAMP_KEY
    CASE_ID_KEY = constants.PARAMETER_CONSTANT_CASEID_KEY
    RESOURCE_KEY = constants.PARAMETER_CONSTANT_RESOURCE_KEY
    TABLE_NAME = "table_name"


def apply(
    db: pd.DataFrame, parameters: Optional[Dict[Any, Any]] = None
) -> str:
    """
    Provides a string containing the required database knowledge for DuckDB querying
    (in order for the LLM to produce meaningful queries).

    Parameters
    ---------------
    db
        Database
    parameters
        Parameters of the method, including:
        - Parameters.ACTIVITY_KEY => the attribute of the log to be used as activity
        - Parameters.TIMESTAMP_KEY => the attribute of the log to be used as timestamp
        - Parameters.CASE_ID_KEY => the attribute of the log to be used as case identifier
        - Parameters.RESOURCE_KEY => the attribute of the log containing the resource
        - Parameters.TABLE_NAME => the name of the dataframe/table

    Returns
    ---------------
    db_knowledge
        String containing the required database knowledge.
    """
    if parameters is None:
        parameters = {}

    activity_key = exec_utils.get_param_value(
        Parameters.ACTIVITY_KEY, parameters, xes_constants.DEFAULT_NAME_KEY
    )
    timestamp_key = exec_utils.get_param_value(
        Parameters.TIMESTAMP_KEY,
        parameters,
        xes_constants.DEFAULT_TIMESTAMP_KEY,
    )
    case_id_key = exec_utils.get_param_value(
        Parameters.CASE_ID_KEY, parameters, constants.CASE_CONCEPT_NAME
    )
    resource_key = exec_utils.get_param_value(
        Parameters.RESOURCE_KEY, parameters, xes_constants.DEFAULT_RESOURCE_KEY
    )
    table_name = exec_utils.get_param_value(
        Parameters.TABLE_NAME, parameters, "dataframe"
    )

    descr = (
        """
The underlying database engine is DuckDB.
The table is called """
        + table_name
        + """.
Each row of the table is corresponding to an event, along with its attributes. There is no separate table containing the process variant.
Please consider the following information: the case identifier is called '"""
        + case_id_key
        + """', the activity is stored inside the attribute '"""
        + activity_key
        + """', the timestamp is stored inside the attribute '"""
        + timestamp_key
        + """', the resource is stored inside the attribute '"""
        + resource_key
        + """'.
You should use the EPOCH function of DuckDB to get the timestamp from the date.
The events are already sorted by timestamp, moreover you cannot use ORDER BY inside a concatenation operator.
    """
    )

    return descr



'''
    PM4Py â€“ A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschrÃ¤nkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from pm4py.util.constants import CASE_CONCEPT_NAME
from pm4py.util.constants import (
    PARAMETER_CONSTANT_CASEID_KEY,
    PARAMETER_CONSTANT_ACTIVITY_KEY,
)
from enum import Enum
from pm4py.util import exec_utils
from typing import Optional, Dict, Any, Union, List
import polars as pl
from pm4py.util import constants, xes_constants


class Parameters(Enum):
    CASE_ID_KEY = PARAMETER_CONSTANT_CASEID_KEY
    ACTIVITY_KEY = PARAMETER_CONSTANT_ACTIVITY_KEY
    DECREASING_FACTOR = "decreasingFactor"
    POSITIVE = "positive"


def get_variants_df(
    df: pl.LazyFrame,
    case_id_key: str = CASE_CONCEPT_NAME,
    activity_key: str = xes_constants.DEFAULT_NAME_KEY,
    parameters: Optional[Dict[Any, Any]] = None,
) -> pl.DataFrame:
    """
    Get variants dataframe from a LazyFrame

    Parameters
    ----------
    df
        LazyFrame
    case_id_key
        Case ID column
    activity_key
        Activity column
    parameters
        Parameters

    Returns
    ----------
    variants_df
        DataFrame with variants
    """
    if parameters is None:
        parameters = {}
    
    # Create variants by concatenating activities per case
    variants_df = (
        df.sort([case_id_key, "time:timestamp"])  # Ensure events are ordered
        .group_by(case_id_key, maintain_order=True)
        .agg(
            pl.col(activity_key)
            .str.concat(constants.DEFAULT_VARIANT_SEP)
            .alias("variant")
        )
        .collect()
    )
    
    return variants_df


def apply(
    df: pl.LazyFrame,
    admitted_suffixes: List[str],
    parameters: Optional[Dict[Union[str, Parameters], Any]] = None,
) -> pl.LazyFrame:
    """
    Apply a filter on variants

    Parameters
    -----------
    df
        LazyFrame
    admitted_suffixes
        List of admitted suffixes (to include/exclude)
    parameters
        Parameters of the algorithm, including:
            Parameters.CASE_ID_KEY -> Column that contains the Case ID
            Parameters.ACTIVITY_KEY -> Column that contains the activity
            Parameters.POSITIVE -> Specifies if the filter should be applied including traces (positive=True)
            or excluding traces (positive=False)
            variants_df -> If provided, avoid recalculation of the variants dataframe

    Returns
    -----------
    df
        Filtered LazyFrame
    """
    if parameters is None:
        parameters = {}

    case_id_glue = exec_utils.get_param_value(
        Parameters.CASE_ID_KEY, parameters, CASE_CONCEPT_NAME
    )
    activity_key = exec_utils.get_param_value(
        Parameters.ACTIVITY_KEY, parameters, xes_constants.DEFAULT_NAME_KEY
    )
    positive = exec_utils.get_param_value(
        Parameters.POSITIVE, parameters, True
    )
    
    # Get or calculate variants dataframe
    variants_df = parameters.get("variants_df")
    if variants_df is None:
        variants_df = get_variants_df(df, case_id_glue, activity_key, parameters)

    # Handle suffix formats
    admitted_suffixes = list(admitted_suffixes)
    if isinstance(admitted_suffixes[0], (list, tuple)):
        # Convert list/tuple suffixes to string format
        admitted_suffixes = [
            constants.DEFAULT_VARIANT_SEP.join(x) for x in admitted_suffixes
        ]

    # Filter variants that end with admitted suffixes
    suffix_expr = None
    for suffix in admitted_suffixes:
        if suffix_expr is None:
            suffix_expr = pl.col("variant").str.ends_with(suffix)
        else:
            suffix_expr = suffix_expr | pl.col("variant").str.ends_with(suffix)
    
    matching_cases = variants_df.filter(suffix_expr).select(case_id_glue)

    if positive:
        # Keep cases with matching suffixes
        ret = df.join(matching_cases.lazy(), on=case_id_glue, how="inner")
    else:
        # Keep cases without matching suffixes
        ret = df.join(matching_cases.lazy(), on=case_id_glue, how="anti")

    return ret


'''
    PM4Py â€“ A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschrÃ¤nkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from pm4py.statistics.passed_time.polars.variants import pre, post, prepost
from enum import Enum
from pm4py.util import exec_utils
from typing import Optional, Dict, Any
import polars as pl


class Variants(Enum):
    PRE = pre
    POST = post
    PREPOST = prepost


VERSIONS = {Variants.PRE, Variants.POST, Variants.PREPOST}


def apply(
    lf: pl.LazyFrame,
    activity: str,
    variant=Variants.PRE,
    parameters: Optional[Dict[Any, Any]] = None,
) -> Dict[str, Any]:
    """
    Gets statistics on execution times of the paths to/from the activity

    Parameters
    ------------
    lf
        LazyFrame
    activity
        Activity
    variant
        Variant:
            - Variants.PRE
            - Variants.POST
            - Variants.PREPOST
    parameters
        Possible parameters of the algorithm

    Returns
    -----------
    dictio
        Dictio containing the times from/to the activity
    """
    return exec_utils.get_variant(variant).apply(
        lf, activity, parameters=parameters
    )


'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from abc import ABC
from collections import Counter
from typing import List, Optional, Collection, Any, Tuple, Generic, Dict

from pm4py.util import nx_utils

from pm4py.algo.discovery.inductive.cuts.abc import Cut, T
from pm4py.algo.discovery.inductive.dtypes.im_dfg import InductiveDFG
from pm4py.algo.discovery.inductive.dtypes.im_ds import (
    IMDataStructureUVCL,
    IMDataStructureDFG,
)
from pm4py.objects.dfg import util as dfu
from pm4py.objects.dfg.obj import DFG
from pm4py.objects.process_tree.obj import Operator, ProcessTree
from pm4py.util.compression.dtypes import UVCL


class LoopCut(Cut[T], ABC, Generic[T]):

    @classmethod
    def operator(
        cls, parameters: Optional[Dict[str, Any]] = None
    ) -> ProcessTree:
        return ProcessTree(operator=Operator.LOOP)

    @classmethod
    def holds(
            cls, obj: T, parameters: Optional[Dict[str, Any]] = None
    ) -> Optional[List[Collection[Any]]]:
        """
        Finds a loop cut in the DFG and returns exactly two groups:
        - groups[0]: the 'do' group (start ∪ end activities plus merged components as per checks)
        - groups[1]: a single 'redo' group obtained by merging all remaining non-empty groups

        If no non-empty redo part remains, returns None.
        """
        dfg = obj.dfg
        start_activities = set(dfg.start_activities.keys())
        end_activities = set(dfg.end_activities.keys())
        if len(dfg.graph) == 0:
            return None

        # Initial groups: do-part is start ∪ end; other parts are connected components after removing boundaries
        groups = [start_activities.union(end_activities)]
        for c in cls._compute_connected_components(dfg, start_activities, end_activities):
            groups.append(set(c.nodes))

        # Apply the original reachability/completeness checks
        groups = cls._exclude_sets_non_reachable_from_start(dfg, start_activities, end_activities, groups)
        groups = cls._exclude_sets_no_reachable_from_end(dfg, start_activities, end_activities, groups)
        groups = cls._check_start_completeness(dfg, start_activities, end_activities, groups)
        groups = cls._check_end_completeness(dfg, start_activities, end_activities, groups)

        # Keep only non-empty groups
        groups = [set(g) for g in groups if len(g) > 0]

        # Require at least a do group and something to redo
        if len(groups) <= 1:
            return None

        # Merge all remaining non-empty groups (from the second to the last) into a single redo group
        redo_merged = set()
        for g in groups[1:]:
            redo_merged.update(g)

        return [set(groups[0]), redo_merged]

    @classmethod
    def _check_start_completeness(
        cls,
        dfg: DFG,
        start_activities: Collection[Any],
        end_activities: Collection[Any],
        groups: List[Collection[Any]],
        parameters: Optional[Dict[str, Any]] = None,
    ) -> List[Collection[Any]]:
        i = 1
        while i < len(groups):
            merge = False
            for a in groups[i]:
                if merge:
                    break
                for x, b in dfg.graph:
                    if x == a and b in start_activities:
                        for s in start_activities:
                            if not (a, s) in dfg.graph:
                                merge = True
            if merge:
                groups[0] = set(groups[0]).union(groups[i])
                del groups[i]
                continue
            i = i + 1
        return groups

    @classmethod
    def _check_end_completeness(
        cls,
        dfg: DFG,
        start_activities: Collection[Any],
        end_activities: Collection[Any],
        groups: List[Collection[Any]],
        parameters: Optional[Dict[str, Any]] = None,
    ) -> List[Collection[Any]]:
        i = 1
        while i < len(groups):
            merge = False
            for a in groups[i]:
                if merge:
                    break
                for b, x in dfg.graph:
                    if x == a and b in end_activities:
                        for e in end_activities:
                            if not (e, a) in dfg.graph:
                                merge = True
            if merge:
                groups[0] = set(groups[0]).union(groups[i])
                del groups[i]
                continue
            i = i + 1
        return groups

    @classmethod
    def _exclude_sets_non_reachable_from_start(
        cls,
        dfg: DFG,
        start_activities: Collection[Any],
        end_activities: Collection[Any],
        groups: List[Collection[Any]],
        parameters: Optional[Dict[str, Any]] = None,
    ) -> List[Collection[Any]]:
        for a in set(start_activities).difference(set(end_activities)):
            for x, b in dfg.graph:
                if x == a:
                    group_a, group_b = None, None
                    for group in groups:
                        group_a = group if a in group else group_a
                        group_b = group if b in group else group_b
                    groups = [
                        group
                        for group in groups
                        if group != group_a and group != group_b
                    ]
                    # we are always merging on the do-part
                    groups.insert(0, group_a.union(group_b))
        return groups

    @classmethod
    def _exclude_sets_no_reachable_from_end(
        cls,
        dfg: DFG,
        start_activities: Collection[Any],
        end_activities: Collection[Any],
        groups: List[Collection[Any]],
        parameters: Optional[Dict[str, Any]] = None,
    ) -> List[Collection[Any]]:
        for b in set(end_activities).difference(start_activities):
            for a, x in dfg.graph:
                if x == b:
                    group_a, group_b = None, None
                    for group in groups:
                        group_a = group if a in group else group_a
                        group_b = group if b in group else group_b
                    groups = [
                        group
                        for group in groups
                        if group != group_a and group != group_b
                    ]
                    groups.insert(0, group_a.union(group_b))
        return groups

    @classmethod
    def _compute_connected_components(
        cls,
        dfg: DFG,
        start_activities: Collection[Any],
        end_activities: Collection[Any],
        parameters: Optional[Dict[str, Any]] = None,
    ):
        nxd = dfu.as_nx_graph(dfg)
        [
            nxd.remove_edge(a, b)
            for (a, b) in dfg.graph
            if a in start_activities
            or a in end_activities
            or b in start_activities
            or b in end_activities
        ]
        [nxd.remove_node(a) for a in start_activities if nxd.has_node(a)]
        [nxd.remove_node(a) for a in end_activities if nxd.has_node(a)]
        nxu = nxd.to_undirected()
        return [
            nxd.subgraph(c).copy() for c in nx_utils.connected_components(nxu)
        ]


class LoopCutUVCL(LoopCut[IMDataStructureUVCL]):

    @classmethod
    def project(
        cls,
        obj: IMDataStructureUVCL,
        groups: List[Collection[Any]],
        parameters: Optional[Dict[str, Any]] = None,
    ) -> List[IMDataStructureUVCL]:
        """
        Split the (compressed) log into one 'do' log and N 'redo' logs.
        Traces are segmented into do/redo slices; redo slices are routed to the
        redo log whose activity set overlaps most with the slice.
        """
        do = set(groups[0])
        redo_groups = [set(g) for g in groups[1:]]

        # For quick membership checks
        redo_activities = set().union(*redo_groups) if redo_groups else set()

        do_log = Counter()
        redo_logs = [Counter() for _ in redo_groups]

        for t, card in obj.data_structure.items():
            do_trace: Tuple[Any, ...] = tuple()
            redo_trace: Tuple[Any, ...] = tuple()

            for e in t:
                if e in do:
                    do_trace += (e,)
                    # flush redo if we were inside a redo slice
                    if redo_trace:
                        redo_logs = cls._append_trace_to_redo_log(
                            redo_trace, redo_logs, redo_groups, card
                        )
                        redo_trace = tuple()
                elif e in redo_activities:
                    redo_trace += (e,)
                    # flush do if we were inside a do slice
                    if do_trace:
                        do_log.update({do_trace: card})
                        do_trace = tuple()
                else:
                    # Safety: if an event is in neither do nor any redo group,
                    # flush any current slice and ignore the event
                    if do_trace:
                        do_log.update({do_trace: card})
                        do_trace = tuple()
                    if redo_trace:
                        redo_logs = cls._append_trace_to_redo_log(
                            redo_trace, redo_logs, redo_groups, card
                        )
                        redo_trace = tuple()

            # Flush tail slices
            if redo_trace:
                redo_logs = cls._append_trace_to_redo_log(
                    redo_trace, redo_logs, redo_groups, card
                )
            do_log.update({do_trace: card})  # keep empty do slices, consistent with original

        logs = [do_log] + redo_logs
        return [IMDataStructureUVCL(l) for l in logs]

    @classmethod
    def _append_trace_to_redo_log(
        cls,
        redo_trace: Tuple[Any, ...],
        redo_logs: List[UVCL],
        redo_groups: List[Collection[Any]],
        cardinality: int,
        parameters: Optional[Dict[str, Any]] = None,
    ) -> List[UVCL]:
        """
        Append a redo slice to the best matching redo log (largest activity overlap).
        """
        if not redo_logs:
            return redo_logs

        activities = set(redo_trace)
        overlaps = [
            (i, len(activities.intersection(set(redo_groups[i]))))
            for i in range(len(redo_groups))
        ]
        overlaps.sort(key=lambda x: (x[1], x[0]), reverse=True)
        target = overlaps[0][0]
        redo_logs[target].update({redo_trace: cardinality})
        return redo_logs


class LoopCutDFG(LoopCut[IMDataStructureDFG]):

    @classmethod
    def project(
        cls,
        obj: IMDataStructureUVCL,
        groups: List[Collection[Any]],
        parameters: Optional[Dict[str, Any]] = None,
    ) -> List[IMDataStructureDFG]:
        dfg = obj.dfg
        dfgs = []
        skippable = [False, False]
        for gind, g in enumerate(groups):
            dfn = DFG()
            for a, b in dfg.graph:
                if a in g and b in g:
                    dfn.graph[(a, b)] = dfg.graph[(a, b)]
                if b in dfg.start_activities and a in dfg.end_activities:
                    skippable[1] = True
            if gind == 0:
                for a in dfg.start_activities:
                    if a in g:
                        dfn.start_activities[a] = dfg.start_activities[a]
                    else:
                        skippable[0] = True
                for a in dfg.end_activities:
                    if a in g:
                        dfn.end_activities[a] = dfg.end_activities[a]
                    else:
                        skippable[0] = True
            elif gind == 1:
                for a in g:
                    dfn.start_activities[a] = 1
                    dfn.end_activities[a] = 1
            dfgs.append(dfn)
        return [
            IMDataStructureDFG(InductiveDFG(dfg=dfgs[i], skip=skippable[i]))
            for i in range(len(dfgs))
        ]



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''

from pm4py.objects.ocel.obj import OCEL
from typing import Dict, Any, Optional
import pandas as pd
from pm4py.objects.ocel.util import ocel_consistency
from pm4py.objects.ocel.util import filtering_utils
from pm4py.objects.log.util import dataframe_utils
from pm4py.util import pandas_utils, constants as pm4_constants


def apply(file_path: str, parameters: Optional[Dict[Any, Any]] = None) -> OCEL:
    """
    Imports an OCEL from a SQLite database using Pandas

    Parameters
    --------------
    file_path
        Path to the SQLite database
    parameters
        Parameters of the import

    Returns
    --------------
    ocel
        Object-centric event log
    """
    if parameters is None:
        parameters = {}

    import sqlite3

    conn = sqlite3.connect(file_path)

    events = pd.read_sql("SELECT * FROM EVENTS", conn)
    objects = pd.read_sql("SELECT * FROM OBJECTS", conn)
    relations = pd.read_sql("SELECT * FROM RELATIONS", conn)

    events = dataframe_utils.convert_timestamp_columns_in_df(
        events,
        timest_format=pm4_constants.DEFAULT_TIMESTAMP_PARSE_FORMAT,
        timest_columns=["ocel:timestamp"],
    )

    relations = dataframe_utils.convert_timestamp_columns_in_df(
        relations,
        timest_format=pm4_constants.DEFAULT_TIMESTAMP_PARSE_FORMAT,
        timest_columns=["ocel:timestamp"],
    )

    ocel = OCEL(
        events=events,
        objects=objects,
        relations=relations,
        parameters=parameters,
    )
    ocel = ocel_consistency.apply(ocel, parameters=parameters)
    ocel = filtering_utils.propagate_relations_filtering(
        ocel, parameters=parameters
    )

    return ocel
